package org.eclipse.jdt.internal.compiler.lookup;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import java.util.List;
import java.util.Set;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ast.Annotation;
import org.eclipse.jdt.internal.compiler.ast.Wildcard;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationBinding;
import org.eclipse.jdt.internal.compiler.lookup.BinaryTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.CaptureBinding;
import org.eclipse.jdt.internal.compiler.lookup.InferenceContext;
import org.eclipse.jdt.internal.compiler.lookup.InferenceVariable;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.UnresolvedReferenceBinding;

public class WildcardBinding extends ReferenceBinding {
  public ReferenceBinding genericType;
  
  public int rank;
  
  public TypeBinding bound;
  
  public TypeBinding[] otherBounds;
  
  char[] genericSignature;
  
  public int boundKind;
  
  ReferenceBinding superclass;
  
  ReferenceBinding[] superInterfaces;
  
  TypeVariableBinding typeVariable;
  
  LookupEnvironment environment;
  
  boolean inRecursiveFunction;
  
  public WildcardBinding(ReferenceBinding genericType, int rank, TypeBinding bound, TypeBinding[] otherBounds, int boundKind, LookupEnvironment environment) {
    this.inRecursiveFunction = false;
    this.rank = rank;
    this.boundKind = boundKind;
    this.modifiers = 1073741825;
    this.environment = environment;
    initialize(genericType, bound, otherBounds);
    if (genericType instanceof UnresolvedReferenceBinding)
      ((UnresolvedReferenceBinding)genericType).addWrapper((TypeBinding)this, environment); 
    if (bound instanceof UnresolvedReferenceBinding)
      ((UnresolvedReferenceBinding)bound).addWrapper((TypeBinding)this, environment); 
    this.tagBits |= 0x1000000L;
    this.typeBits = 134217728;
  }
  
  TypeBinding bound() {
    return this.bound;
  }
  
  int boundKind() {
    return this.boundKind;
  }
  
  public TypeBinding allBounds() {
    if (this.otherBounds == null || this.otherBounds.length == 0)
      return this.bound; 
    ReferenceBinding[] allBounds = new ReferenceBinding[this.otherBounds.length + 1];
    try {
      allBounds[0] = (ReferenceBinding)this.bound;
      System.arraycopy(this.otherBounds, 0, allBounds, 1, this.otherBounds.length);
    } catch (ClassCastException classCastException) {
      return this.bound;
    } catch (ArrayStoreException arrayStoreException) {
      return this.bound;
    } 
    return this.environment.createIntersectionType18(allBounds);
  }
  
  public void setTypeAnnotations(AnnotationBinding[] annotations, boolean evalNullAnnotations) {
    this.tagBits |= 0x200000L;
    if (annotations != null && annotations.length != 0)
      this.typeAnnotations = annotations; 
    if (evalNullAnnotations)
      evaluateNullAnnotations(null, null); 
  }
  
  public void evaluateNullAnnotations(Scope scope, Wildcard wildcard) {
    long nullTagBits = determineNullBitsFromDeclaration(scope, wildcard);
    if (nullTagBits == 0L) {
      TypeVariableBinding typeVariable2 = typeVariable();
      if (typeVariable2 != null) {
        long typeVariableNullTagBits = typeVariable2.tagBits & 0x180000000000000L;
        if (typeVariableNullTagBits != 0L)
          nullTagBits = typeVariableNullTagBits; 
      } 
    } 
    if (nullTagBits != 0L)
      this.tagBits = this.tagBits & 0xFE7FFFFFFFFFFFFFL | nullTagBits | 0x100000L; 
  }
  
  public long determineNullBitsFromDeclaration(Scope scope, Wildcard wildcard) {
    long nullTagBits = 0L;
    AnnotationBinding[] annotations = this.typeAnnotations;
    if (annotations != null)
      for (int i = 0, length = annotations.length; i < length; i++) {
        AnnotationBinding annotation = annotations[i];
        if (annotation != null)
          if (annotation.type.hasNullBit(64)) {
            if ((nullTagBits & 0x100000000000000L) == 0L) {
              nullTagBits |= 0x80000000000000L;
            } else if (wildcard != null) {
              Annotation annotation1 = wildcard.findAnnotation(36028797018963968L);
              if (annotation1 != null)
                scope.problemReporter().contradictoryNullAnnotations(annotation1); 
            } 
          } else if (annotation.type.hasNullBit(32)) {
            if ((nullTagBits & 0x80000000000000L) == 0L) {
              nullTagBits |= 0x100000000000000L;
            } else if (wildcard != null) {
              Annotation annotation1 = wildcard.findAnnotation(72057594037927936L);
              if (annotation1 != null)
                scope.problemReporter().contradictoryNullAnnotations(annotation1); 
            } 
          }  
      }  
    if (this.bound != null && this.bound.isValidBinding()) {
      long boundNullTagBits = this.bound.tagBits & 0x180000000000000L;
      if (boundNullTagBits != 0L)
        if (this.boundKind == 2) {
          if ((boundNullTagBits & 0x80000000000000L) != 0L)
            if (nullTagBits == 0L) {
              nullTagBits = 36028797018963968L;
            } else if (wildcard != null && (nullTagBits & 0x100000000000000L) != 0L) {
              Annotation annotation = wildcard.bound.findAnnotation(boundNullTagBits);
              if (annotation == null) {
                TypeBinding newBound = this.bound.withoutToplevelNullAnnotation();
                this.bound = newBound;
                wildcard.bound.resolvedType = newBound;
              } else {
                scope.problemReporter().contradictoryNullAnnotationsOnBounds(annotation, nullTagBits);
              } 
            }  
        } else {
          if ((boundNullTagBits & 0x100000000000000L) != 0L)
            if (nullTagBits == 0L) {
              nullTagBits = 72057594037927936L;
            } else if (wildcard != null && (nullTagBits & 0x80000000000000L) != 0L) {
              Annotation annotation = wildcard.bound.findAnnotation(boundNullTagBits);
              if (annotation == null) {
                TypeBinding newBound = this.bound.withoutToplevelNullAnnotation();
                this.bound = newBound;
                wildcard.bound.resolvedType = newBound;
              } else {
                scope.problemReporter().contradictoryNullAnnotationsOnBounds(annotation, nullTagBits);
              } 
            }  
          if (nullTagBits == 0L && this.otherBounds != null)
            for (int i = 0, length = this.otherBounds.length; i < length; i++) {
              if (((this.otherBounds[i]).tagBits & 0x100000000000000L) != 0L) {
                nullTagBits = 72057594037927936L;
                break;
              } 
            }  
        }  
    } 
    return nullTagBits;
  }
  
  public ReferenceBinding actualType() {
    return this.genericType;
  }
  
  TypeBinding[] additionalBounds() {
    return this.otherBounds;
  }
  
  public int kind() {
    return (this.otherBounds == null) ? 516 : 8196;
  }
  
  public boolean boundCheck(TypeBinding argumentType) {
    int i;
    int length;
    switch (this.boundKind) {
      case 0:
        return true;
      case 1:
        if (!argumentType.isCompatibleWith(this.bound))
          return false; 
        for (i = 0, length = (this.otherBounds == null) ? 0 : this.otherBounds.length; i < length; i++) {
          if (!argumentType.isCompatibleWith(this.otherBounds[i]))
            return false; 
        } 
        return true;
    } 
    return argumentType.isCompatibleWith(this.bound);
  }
  
  public boolean canBeInstantiated() {
    return false;
  }
  
  public List<TypeBinding> collectMissingTypes(List<TypeBinding> missingTypes) {
    if ((this.tagBits & 0x80L) != 0L)
      missingTypes = this.bound.collectMissingTypes(missingTypes); 
    return missingTypes;
  }
  
  public void collectSubstitutes(Scope scope, TypeBinding actualType, InferenceContext inferenceContext, int constraint) {
    org.eclipse.jdt.internal.compiler.lookup.WildcardBinding wildcardBinding1;
    org.eclipse.jdt.internal.compiler.lookup.WildcardBinding actualWildcard;
    org.eclipse.jdt.internal.compiler.lookup.WildcardBinding wildcardBinding2;
    int k;
    org.eclipse.jdt.internal.compiler.lookup.WildcardBinding actuaIntersection;
    int j;
    org.eclipse.jdt.internal.compiler.lookup.WildcardBinding actualIntersection;
    int i;
    int i3;
    int i2;
    int i1;
    int n;
    int m;
    int length;
    int i4;
    if ((this.tagBits & 0x20000000L) == 0L)
      return; 
    if (actualType == TypeBinding.NULL || actualType.kind() == 65540)
      return; 
    if (actualType.isCapture()) {
      CaptureBinding capture = (CaptureBinding)actualType;
      wildcardBinding1 = capture.wildcard;
    } 
    switch (constraint) {
      case 1:
        switch (this.boundKind) {
          case 1:
            switch (wildcardBinding1.kind()) {
              case 516:
                actualWildcard = wildcardBinding1;
                switch (actualWildcard.boundKind) {
                  case 1:
                    this.bound.collectSubstitutes(scope, actualWildcard.bound, inferenceContext, 1);
                    break;
                } 
                break;
              case 8196:
                wildcardBinding2 = wildcardBinding1;
                this.bound.collectSubstitutes(scope, wildcardBinding2.bound, inferenceContext, 1);
                for (i3 = 0, i4 = wildcardBinding2.otherBounds.length; i3 < i4; i3++)
                  this.bound.collectSubstitutes(scope, wildcardBinding2.otherBounds[i3], inferenceContext, 1); 
                break;
            } 
            this.bound.collectSubstitutes(scope, (TypeBinding)wildcardBinding1, inferenceContext, 1);
            break;
          case 2:
            switch (wildcardBinding1.kind()) {
              case 516:
                actualWildcard = wildcardBinding1;
                switch (actualWildcard.boundKind) {
                  case 2:
                    this.bound.collectSubstitutes(scope, actualWildcard.bound, inferenceContext, 2);
                    for (k = 0, i2 = (actualWildcard.otherBounds == null) ? 0 : actualWildcard.otherBounds.length; k < i2; k++)
                      this.bound.collectSubstitutes(scope, actualWildcard.otherBounds[k], inferenceContext, 2); 
                    break;
                } 
                break;
              case 8196:
                break;
            } 
            this.bound.collectSubstitutes(scope, (TypeBinding)wildcardBinding1, inferenceContext, 2);
            break;
        } 
        break;
      case 0:
        switch (this.boundKind) {
          case 1:
            switch (wildcardBinding1.kind()) {
              case 516:
                actualWildcard = wildcardBinding1;
                switch (actualWildcard.boundKind) {
                  case 1:
                    this.bound.collectSubstitutes(scope, actualWildcard.bound, inferenceContext, 0);
                    for (k = 0, i2 = (actualWildcard.otherBounds == null) ? 0 : actualWildcard.otherBounds.length; k < i2; k++)
                      this.bound.collectSubstitutes(scope, actualWildcard.otherBounds[k], inferenceContext, 0); 
                    break;
                } 
                break;
              case 8196:
                actuaIntersection = wildcardBinding1;
                this.bound.collectSubstitutes(scope, actuaIntersection.bound, inferenceContext, 0);
                for (i1 = 0, i4 = (actuaIntersection.otherBounds == null) ? 0 : actuaIntersection.otherBounds.length; i1 < i4; i1++)
                  this.bound.collectSubstitutes(scope, actuaIntersection.otherBounds[i1], inferenceContext, 0); 
                break;
            } 
            break;
          case 2:
            switch (wildcardBinding1.kind()) {
              case 516:
                actualWildcard = wildcardBinding1;
                switch (actualWildcard.boundKind) {
                  case 2:
                    this.bound.collectSubstitutes(scope, actualWildcard.bound, inferenceContext, 0);
                    for (j = 0, n = (actualWildcard.otherBounds == null) ? 0 : actualWildcard.otherBounds.length; j < n; j++)
                      this.bound.collectSubstitutes(scope, actualWildcard.otherBounds[j], inferenceContext, 0); 
                    break;
                } 
                break;
            } 
            break;
        } 
        break;
      case 2:
        switch (this.boundKind) {
          case 1:
            switch (wildcardBinding1.kind()) {
              case 516:
                actualWildcard = wildcardBinding1;
                switch (actualWildcard.boundKind) {
                  case 1:
                    this.bound.collectSubstitutes(scope, actualWildcard.bound, inferenceContext, 2);
                    for (j = 0, n = (actualWildcard.otherBounds == null) ? 0 : actualWildcard.otherBounds.length; j < n; j++)
                      this.bound.collectSubstitutes(scope, actualWildcard.otherBounds[j], inferenceContext, 2); 
                    break;
                } 
                break;
              case 8196:
                actualIntersection = wildcardBinding1;
                this.bound.collectSubstitutes(scope, actualIntersection.bound, inferenceContext, 2);
                for (m = 0, i4 = (actualIntersection.otherBounds == null) ? 0 : actualIntersection.otherBounds.length; m < i4; m++)
                  this.bound.collectSubstitutes(scope, actualIntersection.otherBounds[m], inferenceContext, 2); 
                break;
            } 
            break;
          case 2:
            switch (wildcardBinding1.kind()) {
              case 516:
                actualWildcard = wildcardBinding1;
                switch (actualWildcard.boundKind) {
                  case 2:
                    this.bound.collectSubstitutes(scope, actualWildcard.bound, inferenceContext, 2);
                    for (i = 0, length = (actualWildcard.otherBounds == null) ? 0 : actualWildcard.otherBounds.length; i < length; i++)
                      this.bound.collectSubstitutes(scope, actualWildcard.otherBounds[i], inferenceContext, 2); 
                    break;
                } 
                break;
            } 
            break;
        } 
        break;
    } 
  }
  
  public char[] computeUniqueKey(boolean isLeaf) {
    char[] genericTypeKey = this.genericType.computeUniqueKey(false);
    char[] rankComponent = (String.valueOf('{') + String.valueOf(this.rank) + '}').toCharArray();
    switch (this.boundKind) {
      case 0:
        wildCardKey = TypeConstants.WILDCARD_STAR;
        return CharOperation.concat(genericTypeKey, rankComponent, wildCardKey);
      case 1:
        wildCardKey = CharOperation.concat(TypeConstants.WILDCARD_PLUS, this.bound.computeUniqueKey(false));
        return CharOperation.concat(genericTypeKey, rankComponent, wildCardKey);
    } 
    char[] wildCardKey = CharOperation.concat(TypeConstants.WILDCARD_MINUS, this.bound.computeUniqueKey(false));
    return CharOperation.concat(genericTypeKey, rankComponent, wildCardKey);
  }
  
  public char[] constantPoolName() {
    return erasure().constantPoolName();
  }
  
  public TypeBinding clone(TypeBinding immaterial) {
    return (TypeBinding)new org.eclipse.jdt.internal.compiler.lookup.WildcardBinding(this.genericType, this.rank, this.bound, this.otherBounds, this.boundKind, this.environment);
  }
  
  public String annotatedDebugName() {
    StringBuffer buffer = new StringBuffer(16);
    AnnotationBinding[] annotations = getTypeAnnotations();
    int i, length;
    for (i = 0, length = (annotations == null) ? 0 : annotations.length; i < length; i++) {
      buffer.append(annotations[i]);
      buffer.append(' ');
    } 
    switch (this.boundKind) {
      case 0:
        return buffer.append(TypeConstants.WILDCARD_NAME).toString();
      case 1:
        if (this.otherBounds == null)
          return buffer.append(CharOperation.concat(TypeConstants.WILDCARD_NAME, TypeConstants.WILDCARD_EXTENDS, this.bound.annotatedDebugName().toCharArray())).toString(); 
        buffer.append(this.bound.annotatedDebugName());
        for (i = 0, length = this.otherBounds.length; i < length; i++)
          buffer.append(" & ").append(this.otherBounds[i].annotatedDebugName()); 
        return buffer.toString();
    } 
    return buffer.append(CharOperation.concat(TypeConstants.WILDCARD_NAME, TypeConstants.WILDCARD_SUPER, this.bound.annotatedDebugName().toCharArray())).toString();
  }
  
  public String debugName() {
    return toString();
  }
  
  public TypeBinding erasure() {
    if (this.otherBounds == null) {
      if (this.boundKind == 1)
        return this.bound.erasure(); 
      TypeVariableBinding var = typeVariable();
      if (var != null)
        return var.erasure(); 
      return (TypeBinding)this.genericType;
    } 
    return (this.bound.id == 1) ? this.otherBounds[0].erasure() : this.bound.erasure();
  }
  
  public char[] genericTypeSignature() {
    if (this.genericSignature == null) {
      switch (this.boundKind) {
        case 0:
          this.genericSignature = TypeConstants.WILDCARD_STAR;
          return this.genericSignature;
        case 1:
          this.genericSignature = CharOperation.concat(TypeConstants.WILDCARD_PLUS, this.bound.genericTypeSignature());
          return this.genericSignature;
      } 
      this.genericSignature = CharOperation.concat(TypeConstants.WILDCARD_MINUS, this.bound.genericTypeSignature());
    } 
    return this.genericSignature;
  }
  
  public int hashCode() {
    return this.genericType.hashCode();
  }
  
  public boolean hasTypeBit(int bit) {
    if (this.typeBits == 134217728) {
      this.typeBits = 0;
      if (this.superclass != null && this.superclass.hasTypeBit(-134217729))
        this.typeBits |= this.superclass.typeBits & 0x13; 
      if (this.superInterfaces != null)
        for (int i = 0, l = this.superInterfaces.length; i < l; i++) {
          if (this.superInterfaces[i].hasTypeBit(-134217729))
            this.typeBits |= (this.superInterfaces[i]).typeBits & 0x13; 
        }  
    } 
    return ((this.typeBits & bit) != 0);
  }
  
  void initialize(ReferenceBinding someGenericType, TypeBinding someBound, TypeBinding[] someOtherBounds) {
    this.genericType = someGenericType;
    this.bound = someBound;
    this.otherBounds = someOtherBounds;
    if (someGenericType != null)
      this.fPackage = someGenericType.getPackage(); 
    if (someBound != null)
      this.tagBits |= someBound.tagBits & 0x2000000020100880L; 
    if (someOtherBounds != null)
      for (int i = 0, max = someOtherBounds.length; i < max; i++) {
        TypeBinding someOtherBound = someOtherBounds[i];
        this.tagBits |= someOtherBound.tagBits & 0x2000000000100800L;
      }  
  }
  
  public boolean isSuperclassOf(ReferenceBinding otherType) {
    if (this.boundKind == 2) {
      if (this.bound instanceof ReferenceBinding)
        return ((ReferenceBinding)this.bound).isSuperclassOf(otherType); 
      return (otherType.id == 1);
    } 
    return false;
  }
  
  public boolean isIntersectionType() {
    return (this.otherBounds != null);
  }
  
  public ReferenceBinding[] getIntersectingTypes() {
    if (isIntersectionType()) {
      ReferenceBinding[] allBounds = new ReferenceBinding[this.otherBounds.length + 1];
      try {
        allBounds[0] = (ReferenceBinding)this.bound;
        System.arraycopy(this.otherBounds, 0, allBounds, 1, this.otherBounds.length);
      } catch (ClassCastException classCastException) {
        return null;
      } catch (ArrayStoreException arrayStoreException) {
        return null;
      } 
      return allBounds;
    } 
    return null;
  }
  
  public boolean isHierarchyConnected() {
    return (this.superclass != null && this.superInterfaces != null);
  }
  
  public boolean enterRecursiveFunction() {
    if (this.inRecursiveFunction)
      return false; 
    this.inRecursiveFunction = true;
    return true;
  }
  
  public void exitRecursiveFunction() {
    this.inRecursiveFunction = false;
  }
  
  public boolean isProperType(boolean admitCapture18) {
    if (this.inRecursiveFunction)
      return true; 
    this.inRecursiveFunction = true;
    try {
      if (this.bound != null && !this.bound.isProperType(admitCapture18))
        return false; 
      if (this.superclass != null && !this.superclass.isProperType(admitCapture18))
        return false; 
      if (this.superInterfaces != null)
        for (int i = 0, l = this.superInterfaces.length; i < l; i++) {
          if (!this.superInterfaces[i].isProperType(admitCapture18))
            return false; 
        }  
      return true;
    } finally {
      this.inRecursiveFunction = false;
    } 
  }
  
  TypeBinding substituteInferenceVariable(InferenceVariable var, TypeBinding substituteType) {
    ReferenceBinding[] arrayOfReferenceBinding;
    boolean haveSubstitution = false;
    TypeBinding currentBound = this.bound;
    if (currentBound != null) {
      currentBound = currentBound.substituteInferenceVariable(var, substituteType);
      haveSubstitution |= TypeBinding.notEquals(currentBound, this.bound);
    } 
    TypeBinding[] currentOtherBounds = null;
    if (this.otherBounds != null) {
      int length = this.otherBounds.length;
      if (haveSubstitution)
        System.arraycopy(this.otherBounds, 0, arrayOfReferenceBinding = new ReferenceBinding[length], 0, length); 
      for (int j = 0; j < length; j++) {
        TypeBinding currentOtherBound = this.otherBounds[j];
        if (currentOtherBound != null) {
          currentOtherBound = currentOtherBound.substituteInferenceVariable(var, substituteType);
          if (TypeBinding.notEquals(currentOtherBound, this.otherBounds[j])) {
            if (arrayOfReferenceBinding == null)
              System.arraycopy(this.otherBounds, 0, arrayOfReferenceBinding = new ReferenceBinding[length], 0, length); 
            arrayOfReferenceBinding[j] = (ReferenceBinding)currentOtherBound;
          } 
        } 
      } 
    } 
    int i = haveSubstitution | ((arrayOfReferenceBinding != null) ? 1 : 0);
    if (i != 0)
      return (TypeBinding)this.environment.createWildcard(this.genericType, this.rank, currentBound, (TypeBinding[])arrayOfReferenceBinding, this.boundKind); 
    return (TypeBinding)this;
  }
  
  public boolean isUnboundWildcard() {
    return (this.boundKind == 0);
  }
  
  public boolean isWildcard() {
    return true;
  }
  
  int rank() {
    return this.rank;
  }
  
  public char[] readableName() {
    StringBuffer buffer;
    int i;
    int length;
    int j;
    char[] result;
    switch (this.boundKind) {
      case 0:
        return TypeConstants.WILDCARD_NAME;
      case 1:
        if (this.otherBounds == null)
          return CharOperation.concat(TypeConstants.WILDCARD_NAME, TypeConstants.WILDCARD_EXTENDS, this.bound.readableName()); 
        buffer = new StringBuffer(10);
        buffer.append(this.bound.readableName());
        for (i = 0, j = this.otherBounds.length; i < j; i++)
          buffer.append('&').append(this.otherBounds[i].readableName()); 
        result = new char[length = buffer.length()];
        buffer.getChars(0, length, result, 0);
        return result;
    } 
    return CharOperation.concat(TypeConstants.WILDCARD_NAME, TypeConstants.WILDCARD_SUPER, this.bound.readableName());
  }
  
  public char[] nullAnnotatedReadableName(CompilerOptions options, boolean shortNames) {
    StringBuffer buffer = new StringBuffer(10);
    appendNullAnnotation(buffer, options);
    switch (this.boundKind) {
      case 0:
        buffer.append(TypeConstants.WILDCARD_NAME);
        result = new char[length = buffer.length()];
        buffer.getChars(0, length, result, 0);
        return result;
      case 1:
        if (this.otherBounds == null) {
          buffer.append(TypeConstants.WILDCARD_NAME).append(TypeConstants.WILDCARD_EXTENDS);
          buffer.append(this.bound.nullAnnotatedReadableName(options, shortNames));
        } else {
          buffer.append(this.bound.nullAnnotatedReadableName(options, shortNames));
          for (int i = 0, j = this.otherBounds.length; i < j; i++)
            buffer.append('&').append(this.otherBounds[i].nullAnnotatedReadableName(options, shortNames)); 
        } 
        result = new char[length = buffer.length()];
        buffer.getChars(0, length, result, 0);
        return result;
    } 
    buffer.append(TypeConstants.WILDCARD_NAME).append(TypeConstants.WILDCARD_SUPER).append(this.bound.nullAnnotatedReadableName(options, shortNames));
    int length;
    char[] result = new char[length = buffer.length()];
    buffer.getChars(0, length, result, 0);
    return result;
  }
  
  ReferenceBinding resolve() {
    TypeBinding resolveType;
    int i, length;
    if ((this.tagBits & 0x1000000L) == 0L)
      return this; 
    this.tagBits &= 0xFFFFFFFFFEFFFFFFL;
    BinaryTypeBinding.resolveType((TypeBinding)this.genericType, this.environment, false);
    switch (this.boundKind) {
      case 1:
        resolveType = BinaryTypeBinding.resolveType(this.bound, this.environment, true);
        this.bound = resolveType;
        this.tagBits |= resolveType.tagBits & 0x800L | 0x2000000000000000L;
        for (i = 0, length = (this.otherBounds == null) ? 0 : this.otherBounds.length; i < length; i++) {
          resolveType = BinaryTypeBinding.resolveType(this.otherBounds[i], this.environment, true);
          this.otherBounds[i] = resolveType;
          this.tagBits |= resolveType.tagBits & 0x800L | 0x2000000000000000L;
        } 
        break;
      case 2:
        resolveType = BinaryTypeBinding.resolveType(this.bound, this.environment, true);
        this.bound = resolveType;
        this.tagBits |= resolveType.tagBits & 0x800L | 0x2000000000000000L;
        break;
    } 
    if (this.environment.usesNullTypeAnnotations())
      evaluateNullAnnotations(null, null); 
    return this;
  }
  
  public char[] shortReadableName() {
    StringBuffer buffer;
    int i;
    int length;
    int j;
    char[] result;
    switch (this.boundKind) {
      case 0:
        return TypeConstants.WILDCARD_NAME;
      case 1:
        if (this.otherBounds == null)
          return CharOperation.concat(TypeConstants.WILDCARD_NAME, TypeConstants.WILDCARD_EXTENDS, this.bound.shortReadableName()); 
        buffer = new StringBuffer(10);
        buffer.append(this.bound.shortReadableName());
        for (i = 0, j = this.otherBounds.length; i < j; i++)
          buffer.append('&').append(this.otherBounds[i].shortReadableName()); 
        result = new char[length = buffer.length()];
        buffer.getChars(0, length, result, 0);
        return result;
    } 
    return CharOperation.concat(TypeConstants.WILDCARD_NAME, TypeConstants.WILDCARD_SUPER, this.bound.shortReadableName());
  }
  
  public char[] signature() {
    if (this.signature == null) {
      switch (this.boundKind) {
        case 1:
          return this.bound.signature();
      } 
      return typeVariable().signature();
    } 
    return this.signature;
  }
  
  public char[] sourceName() {
    switch (this.boundKind) {
      case 0:
        return TypeConstants.WILDCARD_NAME;
      case 1:
        return CharOperation.concat(TypeConstants.WILDCARD_NAME, TypeConstants.WILDCARD_EXTENDS, this.bound.sourceName());
    } 
    return CharOperation.concat(TypeConstants.WILDCARD_NAME, TypeConstants.WILDCARD_SUPER, this.bound.sourceName());
  }
  
  public ReferenceBinding superclass() {
    if (this.superclass == null) {
      TypeBinding superType = null;
      if (this.boundKind == 1 && !this.bound.isInterface()) {
        superType = this.bound;
      } else {
        TypeVariableBinding variable = typeVariable();
        if (variable != null)
          superType = variable.firstBound; 
      } 
      this.superclass = (superType instanceof ReferenceBinding && !superType.isInterface()) ? 
        (ReferenceBinding)superType : 
        this.environment.getResolvedType(TypeConstants.JAVA_LANG_OBJECT, null);
    } 
    return this.superclass;
  }
  
  public ReferenceBinding[] superInterfaces() {
    if (this.superInterfaces == null) {
      if (typeVariable() != null) {
        this.superInterfaces = this.typeVariable.superInterfaces();
      } else {
        this.superInterfaces = Binding.NO_SUPERINTERFACES;
      } 
      if (this.boundKind == 1) {
        if (this.bound.isInterface()) {
          int length = this.superInterfaces.length;
          System.arraycopy(this.superInterfaces, 0, this.superInterfaces = new ReferenceBinding[length + 1], 1, length);
          this.superInterfaces[0] = (ReferenceBinding)this.bound;
        } 
        if (this.otherBounds != null) {
          int length = this.superInterfaces.length;
          int otherLength = this.otherBounds.length;
          System.arraycopy(this.superInterfaces, 0, this.superInterfaces = new ReferenceBinding[length + otherLength], 0, length);
          for (int i = 0; i < otherLength; i++)
            this.superInterfaces[length + i] = (ReferenceBinding)this.otherBounds[i]; 
        } 
      } 
    } 
    return this.superInterfaces;
  }
  
  public void swapUnresolved(UnresolvedReferenceBinding unresolvedType, ReferenceBinding resolvedType, LookupEnvironment env) {
    boolean affected = false;
    if (this.genericType == unresolvedType) {
      this.genericType = resolvedType;
      affected = true;
    } 
    if (this.bound == unresolvedType) {
      this.bound = env.convertUnresolvedBinaryToRawType((TypeBinding)resolvedType);
      affected = true;
    } 
    if (this.otherBounds != null)
      for (int i = 0, length = this.otherBounds.length; i < length; i++) {
        if (this.otherBounds[i] == unresolvedType) {
          this.otherBounds[i] = env.convertUnresolvedBinaryToRawType((TypeBinding)resolvedType);
          affected = true;
        } 
      }  
    if (affected)
      initialize(this.genericType, this.bound, this.otherBounds); 
  }
  
  public String toString() {
    StringBuffer buffer;
    int i;
    int length;
    if (hasTypeAnnotations())
      return annotatedDebugName(); 
    switch (this.boundKind) {
      case 0:
        return new String(TypeConstants.WILDCARD_NAME);
      case 1:
        if (this.otherBounds == null)
          return new String(CharOperation.concat(TypeConstants.WILDCARD_NAME, TypeConstants.WILDCARD_EXTENDS, this.bound.debugName().toCharArray())); 
        buffer = new StringBuffer(this.bound.debugName());
        for (i = 0, length = this.otherBounds.length; i < length; i++)
          buffer.append('&').append(this.otherBounds[i].debugName()); 
        return buffer.toString();
    } 
    return new String(CharOperation.concat(TypeConstants.WILDCARD_NAME, TypeConstants.WILDCARD_SUPER, this.bound.debugName().toCharArray()));
  }
  
  public TypeVariableBinding typeVariable() {
    if (this.typeVariable == null) {
      TypeVariableBinding[] typeVariables = this.genericType.typeVariables();
      if (this.rank < typeVariables.length)
        this.typeVariable = typeVariables[this.rank]; 
    } 
    return this.typeVariable;
  }
  
  public TypeBinding unannotated() {
    return hasTypeAnnotations() ? this.environment.getUnannotatedType((TypeBinding)this) : (TypeBinding)this;
  }
  
  public TypeBinding withoutToplevelNullAnnotation() {
    if (!hasNullTypeAnnotations())
      return (TypeBinding)this; 
    AnnotationBinding[] newAnnotations = this.environment.filterNullTypeAnnotations(getTypeAnnotations());
    return (TypeBinding)this.environment.createWildcard(this.genericType, this.rank, this.bound, this.otherBounds, this.boundKind, newAnnotations);
  }
  
  public TypeBinding uncapture(Scope scope) {
    if ((this.tagBits & 0x2000000000000000L) == 0L)
      return (TypeBinding)this; 
    TypeBinding freeBound = (this.bound != null) ? this.bound.uncapture(scope) : null;
    int length = 0;
    TypeBinding[] freeOtherBounds = (this.otherBounds == null) ? null : new TypeBinding[length = this.otherBounds.length];
    for (int i = 0; i < length; i++)
      freeOtherBounds[i] = (this.otherBounds[i] == null) ? null : this.otherBounds[i].uncapture(scope); 
    return (TypeBinding)scope.environment().createWildcard(this.genericType, this.rank, freeBound, freeOtherBounds, this.boundKind, getTypeAnnotations());
  }
  
  void collectInferenceVariables(Set<InferenceVariable> variables) {
    if (this.bound != null)
      this.bound.collectInferenceVariables(variables); 
    if (this.otherBounds != null)
      for (int i = 0, length = this.otherBounds.length; i < length; i++)
        this.otherBounds[i].collectInferenceVariables(variables);  
  }
  
  public boolean mentionsAny(TypeBinding[] parameters, int idx) {
    if (this.inRecursiveFunction)
      return false; 
    this.inRecursiveFunction = true;
    try {
      if (super.mentionsAny(parameters, idx))
        return true; 
      if (this.bound != null && this.bound.mentionsAny(parameters, -1))
        return true; 
    } finally {
      this.inRecursiveFunction = false;
    } 
    this.inRecursiveFunction = false;
    return false;
  }
  
  public boolean acceptsNonNullDefault() {
    return false;
  }
  
  public long updateTagBits() {
    if (!this.inRecursiveFunction) {
      this.inRecursiveFunction = true;
      try {
        if (this.bound != null)
          this.tagBits |= this.bound.updateTagBits(); 
        if (this.otherBounds != null)
          for (int i = 0, length = this.otherBounds.length; i < length; i++)
            this.tagBits |= this.otherBounds[i].updateTagBits();  
      } finally {
        this.inRecursiveFunction = false;
      } 
    } 
    return super.updateTagBits();
  }
}
